package logwire.web.bo.loader;

import logwire.core.bo.annotation.*;
import logwire.core.bo.field.BizArray;
import logwire.core.bo.field.BizMany;
import logwire.core.bo.field.BizText;
import logwire.core.bo.object.BizObject;
import logwire.core.exceptions.ApplicationException;
import logwire.core.meta.model.Model;
import logwire.core.meta.model.fields.*;
import logwire.core.meta.query.IQuery;
import logwire.core.resource.BeanLoader;
import logwire.core.resource.loader.JavaModelBeanLoader;
import logwire.core.support.Util;
import logwire.core.tenant.TenantClassLoader;
import logwire.core.tenant.TenantProject;
import logwire.web.bo.field.parse.FieldParser;
import logwire.web.bo.field.parse.FieldParserUtil;
import org.elasticsearch.common.Strings;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.io.DefaultResourceLoader;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;
import java.util.stream.Collectors;

public class BoModelBeanLoader extends BeanLoader<IQuery, TenantProject> implements JavaModelBeanLoader {

    @Autowired
    LoaderUtil loaderUtil;

    /**
     * load BoModel 的逻辑：
     * 1、先遍历出所有 BizObject 的子类 得到集合
     * 2、筛选出当中使用 @BizModel 的类，构建 Model，字段包含父类的字段
     * 3、将集合中的剩余类的字段添加到父类对应的Model 中
     *
     * @param project  当前租户
     * @param input    待加载的资源
     * @param consumer 处理加载结果
     * @throws Exception
     */
    @Override
    public void load(TenantProject project, TenantProject input, Consumer<IQuery> consumer) throws Exception {
        TenantClassLoader classLoader = project.getTenantBeanContext().getTenantClassLoader();

        //先扫描到所有的 BizObject (BizObject的所有子类)
        ClassPathScanningCandidateComponentProvider provider = loaderUtil.getScanningBeanProvider();
        provider.setResourceLoader(new DefaultResourceLoader(classLoader));
        provider.addIncludeFilter(loaderUtil.getAbstractTypeFilter(BizObject.class));
        Map<Boolean, List<BeanDefinition>> isBizModel = provider.findCandidateComponents("logwire.web")
                .stream()
                .collect(Collectors.groupingBy(beanDefinition -> {
                    try {
                        Class<?> clazz = classLoader.loadClass(beanDefinition.getBeanClassName());
                        BizModel bizModelAnnotation = clazz.getAnnotation(BizModel.class);
                        return bizModelAnnotation != null;
                    } catch (ClassNotFoundException e) {
                        throw new ApplicationException("");
                    }
                }));
        List<BeanDefinition> bizModelList = isBizModel.get(true);
        for (BeanDefinition beanDefinition : bizModelList) {
            Class<?> clazz = classLoader.loadClass(beanDefinition.getBeanClassName());
            //获取类名和注解里的label
            String className = Util.toLowerCaseFirst(clazz.getSimpleName());
            String label = clazz.getAnnotation(BizModel.class).label();
            //独立model
            Model model = initChildModel();
            setName(className,model,label);
            model.setVerboseName(label);
            model.setVerboseNameCode(className);
            model.setIncludeAuditFields(true);
            model.setIncludeVersionField(true);
            model.setIncludeDomainField(true);
            initField(clazz, model, consumer);
            consumer.accept(model);
        }
    }

    /**
     * 初始化model的字段
     * @param clazz 类
     * @param model model
     * @param consumer
     */
    private void initField(Class<?> clazz, Model model, Consumer<IQuery> consumer) {
        Field[] fields = clazz.getDeclaredFields();
        setPrimaryKey(model);
        //当一个表有多个大文本字段，只需要建一次表，将多个大文本字段的值放在一张表中
        Boolean isFirstText = true;
        //遍历所有字段
        for (Field f : fields) {
            if (f.getAnnotation(Column.class) != null) {
                //带有Column注解的字段
                FieldParser parser = FieldParserUtil.getParser(f);
                logwire.core.meta.model.fields.Field field = parser.parse(f.getName(), f);
                model.addField(field);
            } else if (f.getAnnotation(Text.class) != null&&isFirstText) {
                // 带有Text注解的字段 大文本属性会创建一张子表(Model)
                // 子表名为：BO表名_text，一个BO有多个大字段时，使用同一张表
                if (f.getType() != BizText.class) {
                    throw new ApplicationException("@Text 字段类型必须为BizText");
                }
                newTextModel(model, f, consumer);
                isFirstText = false;
            } else if (f.getAnnotation(Array.class) != null) {
                // 带有Array注解的字段 字段类型必须是BizArray<X>, X可以是String/Double/Boolean/Long/Integer/BigDecimal/OffsetDateTime。
                // 数组属性会创建一张子表(Model)，子表名为：BO表名_属性名；包含的字段：
                if (f.getType() != BizArray.class) {
                    throw new ApplicationException("@Array 字段类型必须为BizArray");
                }
                newArrayModel(model, f, consumer);
            } else if (f.getAnnotation(Many.class) != null) {
                // 带有Many或SlaveMany注解的字段 字段类型必须是BizMany<X>, X为BO类型，未指定时可在注解中指定。
                // 主多对多定义会创建一张子表(Model)，子表名为：BO表名_TXT；包含的字段：
                if (f.getType() != BizMany.class) {
                    throw new ApplicationException("@Many 字段类型必须为BizMany");
                }
                newManyModel(model, f, consumer);
            }
        }
    }

    /**
     * 数组字段值存到一个列中，用逗号分隔 字段：
     * owner_id/所属对象ID   (BigInt)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     * idx/顺序 Integer
     *
     * @param f
     * @param consumer
     */
    private void newArrayModel(Model model, Field f, Consumer<IQuery> consumer) {
        Array array = f.getAnnotation(Array.class);
        //新建数组字段表
        Model arrayModel = initChildModel();
        arrayModel.setName(model.getName() + "_" + f.getName());
        arrayModel.setVerboseName(array.label());

        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        arrayModel.addField(ownerId);
        arrayModel.addField(new IntegerField("idx"));
        //通过泛型找到字段类型
        Type actualTypeArguments = ((ParameterizedType) f.getGenericType()).getActualTypeArguments()[0];
        FieldParser parser = FieldParserUtil.getParserByFieldType(actualTypeArguments);
        logwire.core.meta.model.fields.Field valueField = parser.parse("value", f);
        arrayModel.addField(valueField);

        consumer.accept(arrayModel);
    }

    /**
     * 大文本 字段：
     * owner_id/所属对象ID   (BigInt)
     * field_name/字段名 varchar(50)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newTextModel(Model model, Field f, Consumer<IQuery> consumer) {
        Text text = f.getAnnotation(Text.class);
        //新建大文本字段表
        Model textModel = initChildModel();
        textModel.setName(model.getName()+ "_text");
        textModel.setVerboseName(text.label());

        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        textModel.addField(ownerId);
        textModel.addField(new StringField("field_name", 50));
        textModel.addField(new TextField("value"));

        consumer.accept(textModel);
    }

    /**
     * 多对多关联属性 字段：
     * id
     * master_id 主表ID
     * slave_id  从表ID
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newManyModel(Model model, Field f, Consumer<IQuery> consumer) {
        Many many = f.getAnnotation(Many.class);
        //新建多对多关系表
        Model manyModel = initChildModel();
        manyModel.setName(model.getName() + "_TXT");
        manyModel.setVerboseName(many.label());

        manyModel.addField(new AutoField());
        manyModel.addField(new ForeignKeyField("master_id", model.getName()));
        manyModel.addField(new ForeignKeyField("slave_id", many.referBoName()));
        consumer.accept(manyModel);
    }

    private Model initChildModel(){
        Model childModel = new Model();
        setPrimaryKey(childModel);
        childModel.setIncludeAuditFields(false);
        childModel.setIncludeVersionField(false);
        childModel.setIncludeDomainField(false);
        return childModel;
    }

    /**
     * bo默认主键id为雪花id
     * @param model
     */
    private void setPrimaryKey(Model model){
        BigIntegerField id = new BigIntegerField("id");
        id.setPrimaryKey(true);
        model.addField(id);
    }


    private void setName(String className, Model model, String label) {
        if (!Strings.isNullOrEmpty(label)) {
            model.setName(label);
        } else {
            model.setName(className);
        }
    }

    @Override
    public boolean accept(TenantProject input) {
        return true;
    }
}
